/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "ape_stream.hh"
#include "lcl_array.hh"
#include "DVecType.hh"
#include "jsonStack.hh"
#include "GDSPlot.hh"
#include "html/document.hh"
#include "html/image.hh"
#include "html/Table2C.hh"
#include "html/writer.hh"
#include "TCanvas.h"
#include <cmath>
#include <cstring>
#include <algorithm>
#include <iostream>
#include <sstream>

#include "poissonge.icc"

using namespace std;
typedef triggers::trig_select trig_select;
typedef triggers::trig_veto   trig_veto;

EXECDAT(ape_stream)

//======================================  constructor
ape_stream::ape_stream(int argc, const char* argv[]) 
: DatEnv(argc, argv), _logCL(0), _logCL1D(0), _corr_limit(0.5), _M(16),
   _osc(getDacc())
{
   _config = argv[argc-1];
   configure(_config);
   _trig.set_debug(Debug());
   getDacc().setStride(_stride);
}

//======================================  Destructor
ape_stream::~ape_stream(void) {
}

//======================================  Calculate everything.
void
ape_stream::calculate(void) {
   _stat1d_vector.clear();
   _stat_vector.clear();
   
   //-----------------------------------  Get limits perform 1D tests
   size_t N = _channel.size();
   for (size_t i=0; i<N; i++) {
      if (_channel[i].empty()) {
	 cout << _channel[i].name() << " is empty!" << endl;
	 _channel[i].set_reject(1);
         continue;
      }
      _channel[i].set_limits(_M);

      //--------------------------------  Find worst entry
      channels::chan_stat stat(i);
      bool fail = _channel[i].test1D(_trig.series(), _logCL1D, stat,
				     _reject_outliers);
      if (fail) {
         _stat1d_vector.push_back(stat);
      }
   }
   std::sort(_stat1d_vector.begin(), _stat1d_vector.end());

   //-----------------------------------  get the expected number per cell.
   if (Debug()) cout << "start 2D tests..." << endl;
   lcl_array<int> kArray(_M*_M);
   lcl_array<int> tArray(_M*_M);
   for (size_t i=1; i<N; i++) {
      if (_channel[i].test_reject(_reject_outliers)) continue;
      DVectI dvi(*_channel[i].ref_rank().refDVect());
      size_t _nEnt = dvi.size();
      double trigAvg = _trig.average();
      for (size_t j=0; j<i; j++) {
	 if (_channel[j].test_reject(_reject_outliers)) continue;
	 //cout << "computing {" << _channel[i].name() << ", "
	 //     << _channel[j].name() << "}" << endl;

	 //-----------------------------  calculate matrices
	 calc_matrices(i, j, tArray.get(), kArray.get());
	 
	 //-----------------------------  flag stats
	 stat_2d stat(i, j);
	 calc_stats(tArray.get(), kArray.get(), stat);

         if (stat.correlation < _corr_limit && stat.min_likelihood < _logCL) {
	    size_t savij = stat.min_index;
            size_t kj = savij % _M;
            size_t ki = savij / _M;
	    if (Debug() > 1) {
	       cout << _channel[i].name() << "[" << ki << "] " 
		    << _channel[j].name() << "[" << kj << "] N="
		    << tArray[savij]
		    << " expect=" << kArray[savij]*trigAvg 
		    << " pois=" << stat.min_likelihood << endl;
	       cout << "correlation = " << stat.correlation << endl;
	    }
	    _stat_vector.push_back(stat);
         }
      }
   }

   //------------------------------------  sort results
   std::sort(_stat_vector.begin(), _stat_vector.end());
   lcl_array<bool> used(N);
   for (size_t i=0; i<N; i++) used[i] = false;
   size_t inxok = 0;
   for (size_t i=0; i<_stat_vector.size(); i++) {
      size_t inxa = _stat_vector[i].channel_a;
      size_t inxb = _stat_vector[i].channel_b;
      if (used[inxa] || used[inxb]) continue;
      if (i != inxok)  _stat_vector[inxok] =  _stat_vector[i];
      inxok++;
      used[inxa] = true;
      used[inxb] = true;
   }
}

//======================================  Formatting utility
inline string
CLstring(double logCL) {
   ostringstream msg;
   msg << logCL << " (log " << exp(logCL) << ")";
   return msg.str();
}

inline const char* 
boolString(bool yn) {
   return yn ? "true" : "false";
}

//======================================  Read the configuration file
void
ape_stream::configure(const std::string& cfile) {
   jsonStack js(cfile);
   double CL = 0, CL1D = 0;

   //-----------------------------------  Time parameters
   double ticks = 256;
   js.fetch_data("stride", ticks);
   _stride = ticks;
   ticks = 128;
   js.fetch_data("overlap", ticks);
   _overlap = ticks;

   //-----------------------------------  Numeric parameters
   js.fetch_data("confidence", CL);
   js.fetch_data("log-confidence", _logCL);
   js.fetch_data("confidence-1d", CL1D);
   js.fetch_data("log-confidence-1d", _logCL1D);
   js.fetch_data("correlation-limit", _corr_limit);
   js.fetch_data("reject-outliers",   _reject_outliers);

   //-----------------------------------  OSC info
   js.fetch_data("osc-file", _oscfile);
   js.fetch_data("osc-cond", _osc_cond);

   //-----------------------------------  html directory
   js.fetch_data("html-directory", _html_directory);   

   //-----------------------------------  Trigger selections.
   js.fetch_data("trigger-path", _trig_path);

   string table_name = "sngl_burst";
   js.fetch_data("trigger-table-name", table_name);
   _trig.set_table_name(table_name);

   string time_field = "peak_time";
   js.fetch_data("trigger-time-field", time_field);
   _trig.set_time_field(time_field);

   string time_ns_field = "peak_time_ns";
   js.fetch_data("trigger-time-ns-field", time_ns_field);
   _trig.set_time_ns_field(time_ns_field);

   bool no_retrigger = true;
   js.fetch_data("inhibit-retrigger", no_retrigger);
   _trig. set_no_retrigger(no_retrigger);
   
   if (js.push_member("trigger-select")) {
      jsonStack::autopop pop_tsel(js);
      while (js.iterate()) {
	 jsonStack::autopop pop_tsiter(js);
         static const char* ts_fields[] = {"field", "string-eq", 
					   "num-ge", "num-lt", 0};
	 string inv_fields = js.validate(ts_fields);
	 if (!inv_fields.empty()) {
	    cerr << "Invalid trigger-select field(s); " << inv_fields << endl;
	    continue;
	 }
         trig_select ts;
         js.fetch_data("field", ts.field);
         if (js.fetch_data("string-eq", ts.string)) {
	    ts.type = trig_select::s_string;
         } else if (js.fetch_data("num-ge", ts.number)) {
	    ts.type = trig_select::s_num_ge;
         } else if (js.fetch_data("num-lt", ts.number)) {
	    ts.type = trig_select::s_num_lt;
         } else {
	    ts.type = trig_select::s_pass;
         }
         _trig.add_select(ts);
      }
   }

   if (js.push_member("vetos")) {
      while (js.iterate()) {
         trig_veto tv;
	 tv.val_min = -1.0/0.0;
	 tv.val_max =  1.0/0.0;
         js.fetch_data("channel", tv.channel);
         js.fetch_data("val-min", tv.val_min);
         js.fetch_data("val-max", tv.val_max);
	 _trig.add_veto(tv);
         js.pop();
      }
      js.pop();
   }

   //-----------------------------------  Channel list
   std::vector<std::string> chan_vect;
   size_t nChan = js.fetch_data("channels", chan_vect);
   if (!nChan) throw runtime_error("ape_stream: Error reading channel list");
   for (size_t i=0; i<nChan; i++) {
      if (chan_vect[i].empty()) continue;
      _channel.push_back(channels(chan_vect[i]));
      _channel.back().set_debug(Debug());
      getDacc().addChannel(chan_vect[i]);
   }

   //------------------------------------  Set stuff up
   if (!_oscfile.empty()) _osc.readConfig(_oscfile);
   if (_logCL == 0) {
      if (CL <= 0) CL = 0.002;
      _logCL = log(CL);
   }
   if (_logCL1D == 0) {
      if (CL1D <= 0) CL1D = 0.0002;
      _logCL1D = log(CL1D);
   }
   
   //-----------------------------------  Print
   print_config(cout);
}

//=======================================  Print configuration parameters
void
ape_stream::master_reset(void) {
   for (auto p=_channel.begin(); p != _channel.end(); ++p) {
      p->reset();
   }
}

//=======================================  Print configuration parameters
void
ape_stream::print_config(ostream& out) const {
   out << "stride                 "   << double(_stride) << endl;
   out << "overlap                "   << double(_overlap) << endl;   
   out << "confidence:            "   << CLstring(_logCL) << endl;
   out << "confidence-1d:         "   << CLstring(_logCL1D) << endl;
   out << "reject-outliers:       "   << boolString(_reject_outliers) << endl;
   out << "html-directory:        \"" << _html_directory << "\"" << endl;
   out << "osc-file:              \"" << _oscfile << "\"" << endl;
   out << "osc-cond:              \"" << _osc_cond << "\"" << endl;
   out << "number of channels:    "   << _channel.size() << endl;
   out << "trigger-path:          \"" << _trig_path[0] << "\"" << endl;
   for (size_t i=1; i<_trig_path.size(); i++) {
      out << "                       \"" << _trig_path[i] << "\"" << endl;
   }
   out << "trigger-table-name:    \"" << _trig.table_name() << "\"" << endl;
   out << "trigger-time-field:    \"" << _trig.time_field() << "\"" << endl;
   out << "trigger-time-ns-field: \"" << _trig.time_ns_field() << "\"" << endl;
   out << "inhibit-retrigger:     \"" << boolString(_trig.no_retrigger()) 
       << "\"" << endl;
   out << "trigger selections:    "   << _trig.select_string() << endl;
   out << "trigger vetoes:        "   << _trig.veto_list().size() << endl;
}

//======================================  Calculate matrices
void 
ape_stream::calc_matrices(size_t i, size_t j, int* trig, int* counts) const {
   //-----------------------------------  Zero statistics arrays
   memset(counts, 0, _M*_M*sizeof(int));
   memset(trig,   0, _M*_M*sizeof(int));

   //-----------------------------------  fill matrices
   const DVectI dvi(*_channel[i].ref_rank().refDVect());
   const DVectI dvj(*_channel[j].ref_rank().refDVect());
   const DVectI& dvt = dynamic_cast<const DVectI&>(*_trig.series().refDVect());
   size_t nEnt = dvi.size();
   for (size_t k=0; k<nEnt; k++) {
      size_t ivsj = dvi[k] * _M + dvj[k];
      counts[ivsj]++;
      trig[ivsj] += dvt[k];
   }
}
   
//======================================  Calculate statistics
void 
ape_stream::calc_stats(const int* trig, const int* counts, stat_2d& s) const {
   double trigAvg = _trig.average();
   int savij = -1;
   double wx = 0;
   double ssq=0;
   for (size_t inxij=0; inxij < _M * _M; inxij++) {
      ssq += pow(counts[inxij], 2);
      double xpect = counts[inxij] * trigAvg;
      double pois  = log(poissonge(xpect, trig[inxij]));
      if (pois < wx) {
	 wx = pois;
	 savij = inxij;
      }
   }
   size_t nEnt = _trig.series().getNSample();
   s.min_likelihood = wx;
   if (savij >= 0) {
      s.correlation = sqrt(_M * ssq / (nEnt *nEnt));
      s.min_index   = savij;
      s.trigs_found = trig[savij];
      s.trigs_expected = counts[savij] * trigAvg;
   }
}
   
//======================================  Make the html output
void
ape_stream::results(ostream& out) const {
   double trigAvg = _trig.average();
   out << "----------------------------------------------------------" << endl;
   out << "start time:             " << _trig.series().getStartTime().getS()
       << endl;
   out << "duration:              " << _trig.series().getInterval() << endl;
   out << "outlier channels:      " << _stat1d_vector.size() << endl;
   out << "number triggers:       " << _trig.total() << endl;

   //-------------------------------------  Build 1d results table
   if (!_stat1d_vector.empty()) {
      out << "Single Channel Outliers" << endl;
      out << "Channel   Bin   # found/expected  log(Likelihood)" << endl;
      for (size_t i=0; i < _stat1d_vector.size(); i++) {
	 const channels::chan_stat& s(_stat1d_vector[i]);
	 int id = s.channel_id;
	 out << _channel[id].name()
	     << "  " << _channel[id].bin_id(s.min_index)
	     << "  " << s.trigs_found << "/" << s.trigs_expected
	     << "  " << s.min_likelihood << endl;
      }
   }

   //-----------------------------------  Loop over plots.
   size_t N = _stat_vector.size();
   if (N) {
      if (N > 50) N = 50;
      out << endl << "2D Outliers" << endl;
      for (size_t i=0; i<N; i++) {
	 const stat_2d& stat = _stat_vector[i];
	 size_t chan_a = stat.channel_a;
	 size_t chan_b = stat.channel_b;
	 out << "channel name A    " << _channel[chan_a].name() << endl;
	 out << "channel A bin     " << 
	    _channel[chan_a].bin_id(stat.min_index/_M) << endl;
	 out << "channel name B    " << _channel[chan_b].name() << endl;
	 out << "channel B bin     " << 
	    _channel[chan_b].bin_id(stat.min_index%_M) << endl;
	 out << "correlation       " << stat.correlation << endl;
	 out << "Triggers found    " << stat.trigs_found << endl;
	 out << "Triggers expected " << stat.trigs_expected << endl;
	 out << "log-likelihood    " << stat.min_likelihood << endl;
	 out << endl;
      }
   }
}

//======================================  Collect data from specified channels
void
ape_stream::ProcessData(void) {
   Time t0 =  getDacc().getFillTime();
   cout << "read data at: " << t0 << endl;
   if (!_osc_cond.empty() && !_osc.satisfied(_osc_cond)) return;
   
   size_t N = _channel.size();
   Time tStart, tEnd;
   bool fail = false;
   for (size_t i=0; i<N; i++) {
      const string& name(_channel[i].name());
      const TSeries* ts = getDacc().refData(name);
      if (!ts) {
         cerr << "ape_stream: no data for channel: " << name <<endl;
         continue;
      }
      if (t0 != _channel[i].ref_values().getEndTime()) _channel[i].reset();
      _channel[i].append(*ts);
      if (!tStart) {
	 tStart = _channel[i].ref_values().getStartTime();
	 tEnd =   _channel[i].ref_values().getEndTime();
      } else if (tStart != _channel[i].ref_values().getStartTime()) {
	 cout << "ape_stream: Channel " << name << " timing is inconsistent"
	      << endl;
	 fail = true;
      } else if (tEnd   != _channel[i].ref_values().getEndTime()) {
	 cout << "ape_stream: Channel " << name << " timing is inconsistent"
	      << endl;
	 fail = true;
      }
   }
   
   //-----------------------------------  check for failure - bail if so.
   if (fail) {
      master_reset();
      return;
   }
   
   //-----------------------------------  Full stride? process!
   if (tEnd - tStart >= _stride) {
      if (Debug()) cout << "Starting post-processing for GPS "
			<< tStart.getS() << endl;
      gps_t start = tStart.getS();
      gps_t stop  = tEnd.getS();
      Interval dt = _channel[0].ref_values().getTStep();
      size_t nBin = double(stop - start) / double(dt);

      //--------------------------------  read triggers
      _trig.read_triggers(_trig_path, start, stop, dt);

      //--------------------------------  Veto triggers based on channel values
      if (Debug()) cout << "Vetoing triggers" << endl;
      size_t Nc = _channel.size();
      for (size_t i=0; i<Nc; i++) {
	 _trig.veto_triggers(_channel[i].name(), _channel[i].ref_values());
      }
      _trig.veto_stats();
     
      if (_trig.total()) {
	 if (Debug()) cout << "Starting channel pair test" << endl;
	 calculate();
	 if (!_stat1d_vector.empty() || !_stat_vector.empty()) {
	    results(cout);
	 }
      }
     
      //-----------------------------  Flush to the start of the overlap
      if (tStart + _overlap < tEnd) {
	 Interval flush_delta_t = (tEnd - tStart) - _overlap;
	 for (size_t i=0; i<N; ++i) {
	    _channel[i].next(flush_delta_t);
	 }
      }
      getDacc().setStride(_stride - _overlap);
   } else {
      getDacc().setStride(_stride - Interval(tEnd - tStart));
   }
}
