/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef DQ_MODULE_HH
#define DQ_MODULE_HH 1

#include "MultiStream.hh"
#include "DQ_Bit.hh"
#include "frame_name.hh"
#include "FrWriter.hh"
#include "TSeries.hh"
#include "FilterBase.hh"
#include <vector>
#include <string>
#include <iosfwd>

/**  The %DQ_Module class implements the %DQ_Module monitor. This montor calls 
  *  a boolean-valued c++ function of one or more time series and uses the 
  *  result value to set a specified bit in a uint32_t bit-mask channel. The 
  *  monitor loads a plugin.
  */
class DQ_Module : public MultiStream {
public:
   class filter_template;
   class process_channel;
public:
   /**  %DQ_Module monitor constructor. The monitor requires a single json
     *  configuration file path, and also parses the usual MultiStream 
     *  command line arguments.
     */
   DQ_Module(int argc, const char* argv[]);

   /**  %DQ_Module destructor.
     */
   ~DQ_Module(void);

   /** discard write buffer data up to the specified time.
    */
   void align_buffer(const Time& t);
   
   /**  Collect all the channel names requested as input to the dq bit 
     *  evaluation functions and the channels requested to be copied to
     *  the output frame.
     */ 
   void book_channels(void);

   /**  Find a TSeries in the processed channel list, the dq vector channel 
     *  list, or raw-data channel and return the location: 0 if not found or
     *  1 ... n for other sources.
     *  \brief Locate an operand channel
     *  \param chan channel name string.
     *  \return channel type 0 - n.
     */
   int chan_type(const std::string& chan) const;

   /**  Find a TSeries in the Processed channel list, the dq vector channel 
     *  list, or raw-data channel. An empty time series reference is 
     *  returned if the requested channel is not found.
     *  \brief Find an operand channel
     *  \param chan channel name string.
     *  \return constant reference to a channel.
     */
   const TSeries& find(const std::string& chan) const;
   
   /**  Call the init methods for each of the defined flag instances.
     *  \brief Initialize flag evaluation.
     */
   void init_all(void);
   
   /**  Parse the json configuration file.
     */
   void config(const std::string& cfg_file);

   /**  Flush out any channels with queued data.
    */
   void flush(void);
   
   /**  Process a stride of input data.
     */
   void ProcessData(void);

   /** print out the configuration information.
    */
   void put_config(std::ostream& out) const;

   /**  Reset all the filters and output channels in preparation for a new
     *  non-contiguous segment of input data.
     */
   void reset(void);

   /**  Set the flag specified by the argument DQ_bit structure for the 
     *  specified time slice.
     *  \brief Set a flag.
     *  \param dq_def DQ-bit structure defining the flag.
     *  \param t0     Start of time slice.
     *  \param dt     Length of time slice.
     */
   void set_flag(DQ_bit* dq_def, const Time& t0, Interval dt);

   /**  Buffer output data and write all buffered data to frames ending on the 
     *  stride alignment. If \a all is true, the a final frame is written that
     *  contains all the remaining available data. 
     *  \brief Buffer and write available data
     *  \param all Write data past stride boundary.
     */
   void write_available(bool all=false);
   
   /**  Write a frame containing data from the specified time slice. If the 
     *  writer was not opened previously, it is opened with a name 
     *  corresponding to a file with the slice time and the greater of the 
     *  frame length and the end of an alignment slice. The frame is then 
     *  written to the output file and deleted. If the frame end-time 
     *  corresponds to a multiple of the alignment time or if the last_in_file
     *  argument is true, the file is closed. The file is renamed if the length
     *  is not as expected.
     *  \brief Write a frame.
     *  \param t0    Frame start time
     *  \param dT    Frame length.
     *  \param force if true, the file is written unconditionally.
     */
   void write_frame(const Time& t0, Interval dT, bool force=false);

   //===================================  Internal data structures.
private:
   typedef std::vector<process_channel> process_list;
   typedef process_list::iterator       process_iter;
   
   typedef std::map<std::string, filter_template> filter_map;
   typedef filter_map::value_type filter_map_node;

   /** Find a named channel in the processed channel list.
    */
   process_iter find_process_channel(const std::string& chan);

private:
   double stride;            ///<  computation stride (frame length)
   double flag_step;         ///<  mask calculation stride length
   double timeout;
   int    file_alignment;    ///<  aligment of file times
   int    parity_bit;        ///< Parity bit number or no parity if \< 0
   int    print_config;
   bool   invert;
   frame_name frame_path;
   std::string compression;
   std::string config_file;
   std::string config_version;
   uint32_t    config_crc32;
   std::string history_string;
   //std::vector<std::string> copy_channels;
   std::vector<std::string> copy_history;
   Time     startGPS;
   Time     endGPS;
   std::vector<DQ_bit*> flag_vect;
   FrWriter fr_writer;
   Time     output_start;
   Time     current;
   Time     write_time;
   Time     file_start;
   size_t   frames_out;
   filter_map filterDB;
   process_list     proc_list;
};

//======================================  Internal filter database class
class DQ_Module::filter_template {
public:
   enum filter_enum {kf_none, kf_pipe, kf_ypipe};
   filter_template(void);
   filter_template(const std::string& name, const FilterBase& base,
		   filter_enum type);
   filter_template(const filter_template& ft);
   filter_template& operator=(const filter_template& ft);
   ~filter_template(void);
   TSeries apply(const TSeries& x);
   TSeries apply(const TSeries& x, const TSeries& y);
   const FilterBase* get(void) const;
   FilterBase* get(void);
   const std::string& name(void) const;
   void reset(void);
   std::string type(void) const;
private:
   std::string _name;
   std::unique_ptr<FilterBase> _filter;
   filter_enum _type;
};

//=======================================  Processed channel class
/**  Process channel control block.
 */
class DQ_Module::process_channel {
public:
   /** Enumerate the processing modes.
     */
   enum process_enum {pk_undefined,
		      pk_filter,
		      pk_copy,
		      pk_dqmask,
		      pk_dqvalue};

   /**  Enumerate the process modifier flags.
    */
   enum process_flag {pf_write,
		      pf_last};
   
public:
   /**  Construct a processed channel status class instance.
     *  \brief Processed chanel status class constructor.
     *  \param chan channel name.
     */
   process_channel(const std::string& chan, process_enum ptyp=pk_undefined);

   /**  Destroy a processed channel status class instance.
     *  \brief Processed chanel status class destructor.   
     */
   ~process_channel(void);

   void apply(const TSeries& x);
   void apply(const TSeries& x, const TSeries& y);

   /**  Append the output series to the write buffer, and clear the output
     *  series. Return a reference to the write buffer series.
     */
   TSeries& buffer(void);
   const std::string& channel_x(void) const;
   const std::string& channel_y(void) const;
   void dump(std::ostream&  out) const;
   std::string dump_flags(void) const;

   /**  Add the numbered bit to the enable mask (i.e. the mask of all
     *  bits set by the defined dq flags.
     */
   void enable(int bitno);

   const filter_template& filter(void) const;
   filter_template& filter(void);

   /**  Flush any data available from filter histories and process the remaining
     *  internal channels. buffer the data and write it out.
     */ 
   void flush(void);

   /**  Name of internal processed channel.
    */
   const std::string& name(void) const;

   /**  Preset the DQ bit channel time series to all trigger bits OK. The 
     *  evaluated flags are then "OR"ed with the bit-masks.
     *  \brief Preset the time series.
     *  \param t0 Start time for the current stride.
     *  \param dT flag sample interval
     *  \param N  Number of time-series samples in this stride.
     */
   void preset(const Time& t0, Interval dT, size_t N);

   process_enum process_type(void) const;
   const TSeries& ref_series(void) const;
   TSeries& ref_series(void);
   void reset(void);
   void set_channel_x(const std::string& xname);
   void set_channel_y(const std::string& yname);
   void set_filter(const filter_template& filt);

   /**  Set the process modifier flag to the specified state.
    */
   void set_process_flag(process_flag flag, bool set=true);
   void set_process_type(process_enum mode);
   void set_process_type(const std::string& mode);

   /**  Test the specified process modifier flag state.
    */
   bool test_process_flag(process_flag flag) const;
   std::string type(void) const;
   bool write(void) const;

private:
   process_enum    _process_type;
   uint32_t        _process_flags;
   std::string     _chan_name;
   std::string     _chan_inx;
   std::string     _chan_iny;
   filter_template _filt;
   uint32_t        _enable_mask;
   TSeries         _output;
   TSeries         _write_buffer;
};


//======================================  Inline methods

//====================================== filter_template
inline const FilterBase*
DQ_Module::filter_template::get(void) const {
   return _filter.get();
}

inline FilterBase*
DQ_Module::filter_template::get(void) {
   return _filter.get();
}

inline const std::string&
DQ_Module::filter_template::name(void) const {
   return _name;
}

//======================================  process_channel
inline const std::string& 
DQ_Module::process_channel::channel_x(void) const {
   return _chan_inx;
}

inline const std::string&
DQ_Module::process_channel::channel_y(void) const {
   return _chan_iny;
}

inline void
DQ_Module::process_channel::enable(int bitno) {
   _enable_mask |= (1<<bitno);
}

inline const DQ_Module::filter_template&
DQ_Module::process_channel::filter(void) const {
   return _filt;
}

inline DQ_Module::filter_template&
DQ_Module::process_channel::filter(void) {
   return _filt;
}

inline DQ_Module::process_iter
DQ_Module::find_process_channel(const std::string& chan) {
   for (auto itr=proc_list.begin(); itr != proc_list.end(); itr++) {
      if (itr->name() == chan) return itr;
   }
   return proc_list.end();
}

inline const std::string&
DQ_Module::process_channel::name(void) const {
   return _chan_name;
}

inline DQ_Module::process_channel::process_enum
DQ_Module::process_channel::process_type(void) const {
   return _process_type;
}

inline const TSeries&
DQ_Module::process_channel::ref_series(void) const {
   return _output;
}

inline TSeries&
DQ_Module::process_channel::ref_series(void) {
   return _output;
}

inline bool 
DQ_Module::process_channel::test_process_flag(process_flag flag) const {
   return ((1 << flag) & _process_flags) != 0;
}

inline bool 
DQ_Module::process_channel::write(void) const {
   return test_process_flag(pf_write);
}

#endif
