/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "gatemonitor.hh"
#include "DaccAPI.hh"
#include "frame_name.hh"
#include "GDSPlot.hh"
#include "jsonStack.hh"
#include "html/align.hh"
#include "html/document.hh"
#include "html/linebreak.hh"
#include "html/link.hh"
#include "html/Table2C.hh"
#include "html/text.hh"
#include "html/writer.hh"
#include <algorithm>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <cstdlib>
#include <unistd.h>

//-->  The next three lines are needed if you are going to generate triggers.
//     The descriptive title in PIDTITLE should the monitor function.
#define PIDCVSHDR "$Header: https://git.ligo.org/gds/Monitors/gating_monitor/gatemonitor.cc 2.19 2022-7-28 22:00:00 john.zweizig@LIGO.ORG $"
#define PIDTITLE  "Gating statistics monitor"
#include "ProcIdent.hh"

EXECDAT(gating_monitor)

using namespace std;

//======================================  push unique value
template<class T>
void
push_unique(std::vector<T>& vec, const T& val) {
   for (auto ith : vec) if (ith == val) return;
   vec.push_back(val);
}

//======================================  Gating monitor constructor
gating_monitor::gating_monitor(int argc,const char* argv[])
   : DatEnv(argc, argv), _stride(1.0), _osc_list(getDacc()),
     _monitor_pfx("GATEM"), _seg_acct(_trig_client)
{
   if (argc < 2) {
      finish();
      cerr << "gating_monitor: No configuration file specified." << endl;
      return;
   }

   try {
      if (!configure(argv[argc-1])) {
         cerr << "gating_monitor: Error in configuration." << endl;
         finish();
         return;
      }
   } catch (exception& e) {
      cerr << "gating_monitor: Exception occurred in configure: "
           << e.what() << endl;
      finish();
      return;
   }
   if (_gates.empty()) {
      finish();
      cerr << "gating_monitor: No gate conditions specified." << endl;
      return;
   }

   //-----------------------------------  Set up the data accessor
   getDacc().setStride(_stride);
   getDacc().setIgnoreMissingChannel(true);
   vector<string> chans;
   for (auto& gc : _gates) {
      if (!gc.gatevec().empty()) push_unique(chans, gc.gatevec());
      if (!gc.thresh().empty())  push_unique(chans, gc.thresh());
   }

   //-----------------------------------  Request the channel names.
   for (auto chname : chans) {
      getDacc().addChannel(chname);
   }

   //-----------------------------------  Set up trender and trend channels
   _trender.setIFO(_ifo);
   _trender.setType(Trend::kMinute);
   _trender.setAutoUpdate(false);
   _trender.setMonID(_monitor_pfx);
   _trender.setName(_trend_name);
   if (!_trend_file.empty()) _trender.setFile(_trend_file);
   string chan_pfx = _ifo + ":DMT-" + _monitor_pfx;
   for (auto& td : _trends) {
      td.trend_setup(_trender, chan_pfx);
   }
}

//======================================  Gating monitor destructor
gating_monitor::~gating_monitor(void) {
   put_html();
}

//======================================  Gating monitor destructor
inline gate_cond_vect::const_iterator
find_gate(const gate_cond_vect& vect, const std::string& name) {
   return std::find(vect.begin(), vect.end(), name);
}


//======================================  Gating monitor destructor
bool
gating_monitor::configure(const std::string& file) {
   bool config_ok = true;
   jsonStack js(file);

   //-----------------------------------  Get the ifo name
   if (getenv("LIGOSITE")) {
      string site = getenv("LIGOSITE");
      if (site == "LHO")      _ifo = "H1";
      else if (site == "LLO") _ifo = "L1";
   } else if (getenv("DMTIFOS")) {
      string ifos = getenv("DMTIFOS");
      if (ifos[0] == 'H')      _ifo = "H1";
      else if (ifos[0] == 'L') _ifo = "L1";
   }
   js.fetch_data("ifo", _ifo);

   //-----------------------------------  Get osc file and condition
   string oscfile;
   if (js.fetch_data("osc-file", _osc_file)) {
      js.fetch_data("osc-cond", _osc_cond);
      _osc_list.readConfig(_osc_file);
   }

   //-----------------------------------  get processing stride
   double stride = 1.0;
   if (js.fetch_data("stride", stride)) {
      _stride = Interval(stride);
   }

   //----------------------------------- statistics intervals
   if (!js.fetch_data("test_intervals", _test_intervals)) {
      double xv[] = {60.0, 300.0, 3600.0, 86400.0};
      _test_intervals.assign(xv, xv+4);
   }

   //-----------------------------------  Trend information
   js.fetch_data("monitor-prefix", _monitor_pfx);
   _trend_name = _monitor_pfx;
   js.fetch_data("trend-name", _trend_name);
   js.fetch_data("trend-file-path", _trend_file);

   //-----------------------------------  parse the gates
   if (js.push_member("gates")) {
      jsonStack::autopop popper(js);
      while (js.iterate()) {
	 jsonStack::autopop popper2(js);

	 string tmp_string("noname");
	 js.fetch_data("name", tmp_string);
	 gate_cond gc(tmp_string);
	 tmp_string.clear();
	 js.fetch_data("bit-channel", tmp_string);
	 int bitno = 0;
	 js.fetch_data("bit-number", bitno);
	 gc.set_gatevec(tmp_string, bitno);
	 bool inv=true;
	 js.fetch_data("bit-inverse", inv);
	 gc.set_invert(inv);
	 double tmp_dble(1.0);
	 if (js.fetch_data("gate-threshold", tmp_dble)) {
	    gc.set_thrcut(tmp_dble);
	 }
	 tmp_string.clear();
	 if (js.fetch_data("thresh-channel", tmp_string)) {
	    gc.set_thresh(tmp_string);
	 }

	 //-----------------------------  Get the plot parameters
	 if (js.push_member("plot")) {
	    plot_pars pp;
	    double xmin(0), xmax(0), ymin(0), ymax(0);
	    if ((js.fetch_data("xmin", xmin) + js.fetch_data("xmax", xmax) +
		 js.fetch_data("ymin", ymin) + js.fetch_data("ymax", ymax))) {
	       pp.set_range(xmin, xmax, ymin, ymax);
	    }
	    string title;
	    if (js.fetch_data("title", title))  pp.title(title);
	    bool btmp;
	    if (js.fetch_data("logbin", btmp)) pp.logbin(btmp);
	    if (js.fetch_data("xlog", btmp))   pp.xlog(btmp);
	    if (js.fetch_data("ylog", btmp))   pp.ylog(btmp);
	    gc.set_plot(pp);
	    js.pop();
	 }

	 //-----------------------------  Add the gate definition
	 gc.set_debug(Debug());
	 _gates.push_back(move(gc));
      }
   }

   //-----------------------------------  parse the segment definitions
   if (js.push_member("segments")) {
      jsonStack::autopop popper(js);
      while (js.iterate()) {
	 jsonStack::autopop popper2(js);
	 string tmpstr;
	 if (!js.fetch_data("name", tmpstr)) {
	    cerr << "gating_monitor::configure: Segment name not specified"
		 << endl;
	    config_ok = false;
	    continue;
	 }
	 segment_def sd(tmpstr, _seg_acct);

	 if (!js.fetch_data("gate",tmpstr)) {
	    cerr << "gating_monitor::configure: Segment gate not specified"
		 << endl;
	    config_ok = false;
	    continue;
	 }
	 if (find_gate(_gates, tmpstr) ==  _gates.end()) {
	    cerr << "gating_monitor::configure: Segment gate \""
		 << tmpstr << "\" not defined." << endl;
	    config_ok = false;
	    continue;
	 }
	 sd.set_gate(tmpstr);

	 if (!js.fetch_data("mode", tmpstr)) {
	    cerr << "gating_monitor::configure: Segment mode not specified"
		 << endl;
	    config_ok = false;
	    continue;
	 }
	 if (!sd.set_mode(tmpstr)) continue;

	 double tmpdbl = 0;
	 if (js.fetch_data("fraction", tmpdbl)) sd.set_fraction(tmpdbl);
	 if (js.fetch_data("average-time", tmpdbl)) sd.set_average_time(tmpdbl);
	 if (js.fetch_data("max-time", tmpdbl)) sd.set_on_time(tmpdbl);

	 _seg_acct.addSegment(sd.name(), 1, "Generated by gating_monitor");
	 _segments.push_back(std::move(sd));
      }
   }

   //-----------------------------------  parse the trend definitions
   if (js.push_member("trends")) {
      jsonStack::autopop popper(js);
      while (js.iterate()) {
         jsonStack::autopop popper2(js);
	 string trendname, trendgate;
	 if (!js.fetch_data("name", trendname)) {
	    cerr << "gatemonitor: trend name not specified" << endl;
	    config_ok = false;
	    continue;
	 }

	 if (!js.fetch_data("gate", trendgate)) {
	    cerr << "gatemonitor: trend gate name not specified" << endl;
	    config_ok = false;
	    continue;
	 }

	 if (find_gate(_gates, trendgate) ==  _gates.end()) {
	    cerr << "gating_monitor::configure: Trend gate \""
		 << trendgate << "\" not defined." << endl;
	    config_ok = false;
	    continue;
	 }

	 trend_def td(trendname, trendgate);
	 bool yorn = false;
	 if (js.fetch_data("enable-dead", yorn))  td.set_enable_dead(yorn);
	 if (js.fetch_data("enable-discr", yorn)) td.set_enable_discr(yorn);
	 string chanpfx;
	 if (js.fetch_data("channel-prefix", chanpfx)) td.set_prefix(chanpfx);
	 _trends.push_back(td);
      }
   }

   return config_ok;
}

//======================================  Gating monitor destructor
void
gating_monitor::ProcessData(void) {
   //-----------------------------------  Keep track of processing time
   Time now = getDacc().getFillTime();
   if (!_start) {
      _start = now;
      _current = _start;
   } else if (_current != now) {
      reset();
   }
   Interval dt = getDacc().getFillStride();
   _current = now +dt;

   //-----------------------------------  Set the ready timeline
   bool ready(false);
   if (_osc_cond.empty()) {
      _ready_times.add_seg(now, dt);
   } else if (_osc_list.satisfied(_osc_cond)) {
      _ready_times.add_seg(now, dt);
      ready = true;
   }
   // cout << "Time: " << now.getS() << " ready: " << ready << endl;

   //----------------------------------  loop over gates
   bool end_debug = Debug() && (_start + Interval(100.0) < _current);
   for (auto i=_gates.begin(); i!= _gates.end(); i++) {
      i->collect(getDacc(), ready);
      if (end_debug) i->set_debug(0);
   }

   //----------------------------------  calculate segments
   _seg_acct.start_stride(now);
   for (auto& sd : _segments) {
      auto gate_j = find_gate(_gates, sd.gate());
      if (gate_j == _gates.end()) {
	 throw logic_error("gating_monitor: gate condition not found");
      }
      sd.seg_gen(_ready_times, gate_j->time_line(), now, dt);
   }
   _seg_acct.update(_current);

   //-----------------------------------  generate the trends
   for (auto& td : _trends) {
      td.trend_stride(now, dt, *find_gate(_gates, td.gate()), _trender);
   }
   _trender.Update(_current);

   //-----------------------------------  generate the html status page
   if ((_current.getS() % 60) == 0) {
      put_html();
   }
}

//======================================  toaster utility
static string
toaster(double x, int p) {
   ostringstream s;
   s << fixed << setprecision(p) << x;
   return s.str();
}

//======================================  Gating monitor destructor
void
gating_monitor::put_html(void) const {
   const char* time_fmt="%s (%3M %d, %Y %02H:%02N:%02S %Z)";
   //-----------------------------------  Global parameter table
   html::document doc;
   html::Table2C* pars = new html::Table2C("");
   pars->addRow("Start Time",    TimeString(_start, time_fmt));
   pars->addRow("Current Time",  TimeString(_current, time_fmt));
   pars->addRow("OSC File",      _osc_file);
   pars->addRow("OSC Condition", _osc_cond);
   pars->addRow("stride",        double(_stride));
   if (!_trends.empty()) {
      pars->addRow("Trend channel group", _trender.getName());
      if (_trender.getFileName().empty()) {
         const char* dmtout = getenv("DMTRENDOUT");
         char cwd[128];
         string tmp = (dmtout) ? dmtout : getcwd(cwd, sizeof(cwd));
         pars->addRow("Trend directory", tmp);
      } else {
         pars->addRow("Trend file name", _trender.getFileName());
      }
   }
   doc.addObject(pars);

   //-----------------------------------  Get live times.
   if (_ready_times.check()) cerr << "Error in ready time line" << endl;
   vector<Interval> live_time(_test_intervals.size());
   for (size_t i=0;  i < _test_intervals.size(); i++) {
      Interval rdt = _test_intervals[i];
      live_time[i] = _ready_times.on_time(_current-rdt, rdt);
   }

   //-----------------------------------  Build the gate definition table
   html::table* gate_tab = new html::table;
   gate_tab->setBorder(true);
   gate_tab->addColumn("name");
   gate_tab->addColumn("bit-vector");
   gate_tab->addColumn("bit-number");
   gate_tab->addColumn("threshold");
   gate_tab->addColumn("threshold-values");
   int plotid = 0;
   for (auto i=_gates.begin(); i != _gates.end(); i++) {
      int irow = gate_tab->addRow();
      gate_tab->insertData(irow, 0, i->name());
      gate_tab->insertData(irow, 1, i->gatevec());
      gate_tab->insertData(irow, 2, long(i->bitno()));
      gate_tab->insertData(irow, 3, i->thrcut());
      if (i->thresh().empty()) {
	 gate_tab->insertData(irow, 4, i->thresh());
      } else {
	 ostringstream file;
	 Time::ulong_t gps = _current.getS();
	 if (!(gps % 1000)) {
	    file << "threshold-" << ++plotid << "-gps-" << gps << ".png";
	 } else {
	    file << "threshold-" << ++plotid << ".png";
	 }
	 string path;
	 if (getenv("DMTHTMLOUT")) {
	    path = getenv("DMTHTMLOUT");
	    path += "/";
	 }
	 path += file.str();
	 GDSPlot plot;
	 i->plot_threshold(plot);
	 plot.print(path);
	 gate_tab->refCell(irow, 4) << html::link(i->thresh(), file.str());
      }
   }
   doc.addObject(gate_tab);

   //-----------------------------------  Build the dead-time table
   html::table* dead_tab = new html::table;
   dead_tab->setBorder(true);
   dead_tab->addColumn("gate name");
   for (size_t i=0; i < _test_intervals.size(); i++) {
      ostringstream msg;
      msg << "last " << long(_test_intervals[i]) << "s";
      dead_tab->addColumn(msg.str());
   }
   for (size_t i=0; i < _test_intervals.size(); i++) {
      dead_tab->refHeader(i+1) << html::align("center")
			       << html::linebreak() << "dead time"
			       << html::linebreak() << "(dead %)";
   }

   //-----------------------------------  Add the live time row.
   int lRow = dead_tab->addRow();
   dead_tab->insertData(lRow, 0, "live");
   for (auto i=0; i < _test_intervals.size(); i++) {
      Interval dt(_test_intervals[i]);
      double live(live_time[i]);
      ostringstream msg;
      msg << fixed << setprecision(4) << live;
      dead_tab->refCell(lRow, i+1) << msg.str();
   }

   for (auto j=0; j < _gates.size(); j++) {
      Interval dt(_test_intervals[j]);
      double live(_ready_times.on_time(_current - dt, dt));
      timeline gate_j = _gates[j].time_line();
      if (gate_j.check()) cerr << "Error in gate " << j << " time line" << endl;
      timeline dead_j = gate_j & _ready_times;
      if (dead_j.check()) cerr << "Error in dead " << j << " time line" << endl;
      int iRow = dead_tab->addRow();
      dead_tab->insertData(iRow, 0, _gates[j].name());
      for (auto i=0; i < _test_intervals.size(); i++) {
	 Interval rdt(_test_intervals[i]);
	 size_t count = dead_j.count(_current - rdt, rdt);
	 double gate  = double(dead_j.on_time(_current-rdt, rdt));
	 double dead  = (!live_time[i]) ? 0.0 :(gate * 100.0) / live_time[i];
	 ostringstream msg;
	 //msg << fixed << setprecision(4) << count << "<br>" << gate
	 //    << "<br>" << "(" << dead << ")";
	 dead_tab->refCell(iRow, i+1)
	    << html::align("center")
	    << html::text(count) << html::linebreak()
	    << html::text(toaster(gate, 4)) << html::linebreak()
	    << (html::text("(") << toaster(dead, 4) << "%)");
      }
   }
   doc.addObject(dead_tab);

   //-----------------------------------  Build the segment table
   html::table* seg_tab = new html::table;
   seg_tab->setBorder(true);
   seg_tab->addColumn("segment name");
   seg_tab->addColumn("gate condition");
   seg_tab->addColumn("mode");
   seg_tab->addColumn("avg time");
   seg_tab->addColumn("max time");
   seg_tab->addColumn("fraction %");
   seg_tab->addColumn("#segments");
   seg_tab->addColumn("deadtime");
   for (auto i=_segments.begin(); i != _segments.end(); i++) {
      int iRow = seg_tab->addRow();
      seg_tab->insertData(iRow, 0, i->name());
      seg_tab->insertData(iRow, 1, i->gate());
      seg_tab->insertData(iRow, 2, segment_def::mode_name(i->mode()));
      seg_tab->insertData(iRow, 3, double(i->avg_time()));
      seg_tab->insertData(iRow, 4, double(i->max_time()));
      seg_tab->insertData(iRow, 5, i->fraction()*100.0, -1, 2);
      seg_tab->insertData(iRow, 6, long(i->count()));
      seg_tab->insertData(iRow, 7, double(i->total()));
   }
   doc.addObject(seg_tab);

   //-----------------------------------  Write the htmldocument
   string html_path=frame_name::resolve_env("$DMTHTMLOUT/index.html");
   ofstream out(html_path);
   if (!out.is_open()) throw runtime_error("Unable to open html output file");
   html::writer wrt(out);
   doc.write(wrt);
}

//======================================  gating monitor reset method
void
gating_monitor::reset(void) {
}
